// Copyright (C) 2024 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:flutter/foundation.dart';
import 'package:rxdart/subjects.dart';
import 'package:wyatt_authentication/src/authentication_platform.dart';
import 'package:wyatt_authentication/src/authentication_provider.dart';
import 'package:wyatt_authentication/src/authentication_status.dart';

typedef Auth = AuthenticationManager;

class AuthenticationManager {
  AuthenticationManager._();

  static final AuthenticationManager _instance = AuthenticationManager._();

  static AuthenticationPlatform<dynamic, dynamic>? platform;

  static final AuthenticationManager instance = _instance;

  final _status = BehaviorSubject.seeded(AuthenticationStatus.unknown);

  final ValueNotifier<Object?> accountNotifier = ValueNotifier(null);

  /// Initialize the authentication manager with the given [platform]
  static void initialize({
    required AuthenticationPlatform<dynamic, dynamic> platform,
  }) {
    AuthenticationManager.platform = platform;

    platform.accountChanges.listen((account) {
      if (account == null) {
        _instance.accountNotifier.value = null;
        _instance._status.add(AuthenticationStatus.unauthenticated);
      } else {
        _instance.accountNotifier.value = account;
        _instance._status.add(AuthenticationStatus.authenticated);
      }
    });
  }

  void _uninitializedPlatform() {
    if (platform == null) {
      throw StateError('AuthenticationPlatform is not initialized');
    }
  }

  /// Sign in with the given [provider]
  Future<void> signIn(AuthenticationProvider provider) async {
    if (provider is! SignInProvider) {
      throw Exception('$provider must be a sign in provider');
    }

    _uninitializedPlatform();

    return platform!.signIn(provider);
  }

  /// Sign up with the given [provider]
  Future<void> signUp(AuthenticationProvider provider) async {
    if (provider is! SignUpProvider) {
      throw Exception('$provider must be a sign up provider');
    }

    _uninitializedPlatform();

    return platform!.signUp(provider);
  }

  /// Sign out
  Future<void> signOut() async {
    _uninitializedPlatform();

    return platform!.signOut();
  }

  /// Stream of the current authentication status
  Stream<AuthenticationStatus> get status => _status.stream;

  /// Current authentication status
  AuthenticationStatus get currentStatus => _status.value;

  /// The current account or null if unauthenticated
  /// [A] is the type of the account
  A? currentAccount<A>() => accountNotifier.value as A?;

  /// Dispose the authentication manager
  /// Call this method when the authentication manager is no longer needed
  /// to free resources
  void dispose() {
    _status.close();
  }
}
